/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2003-2010, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.ellipticFunctions;

import de.jtem.mfc.field.Complex;

/**
 * Class provides Jacobi's elliptic theta functions.
 *
 * @author Markus Schmies, Boris Springborn
 * @version 0.5
 */

public class Elliptic {

  static final double EPS = 1e-12;

  Elliptic() {
  }

  final static Complex ZERO = new Complex(0);

  /**
   * Computes theta constants c0 = &theta;<sub>0</sub>(0,&tau;), c1 = &theta;<sub>1</sub>(0,&tau;) = 0,
   * c2 = &theta;<sub>2</sub>(0,&tau;) = 0, and c3 = &theta;<sub>3</sub>(0,&tau;).
   * The result is stored in the array c.
   * The array may have null entries, which causes the routine to create
   * instances.
   * @param tau lattice paramter &tau;
   * @param eps absoute error for theta functions
   * @param c array with the four theta constants (on output)
   */
  static void thetaConstants(final Complex tau,
                                    final double eps,
                                    final Complex[] c) {
    if (c.length != 4) {
      throw new IllegalArgumentException("there are four constants");
    }

    for (int i = 0; i < 4; i++) {
      if (c[i] == null) {
        c[i] = new Complex();

      }
    }
    c[1].assign(0);

    theta4(ZERO, tau, eps, c[0]);
    theta2(ZERO, tau, eps, c[2]);
    theta3(ZERO, tau, eps, c[3]);
  }

  /**
    * Computes theta constants &theta;<sub>0</sub>(0,&tau;), &theta;<sub>1</sub>(0,&tau;) = 0,
    * &theta;<sub>2</sub>(0,&tau;) = 0, and &theta;<sub>3</sub>(0,&tau;).
    * The result is stored in the array c.
    * The array may have null entries, which causes the routine to create
    * instances.
    * @param tau lattice paramter &tau;
    * @param c array with the four theta constants (on output)
    */
   public static void thetaConstants(final Complex tau,
                                     final Complex[] c) {
     if (c.length != 4) {
       throw new IllegalArgumentException("there are four constants");
     }

     for (int i = 0; i < 4; i++) {
       if (c[i] == null) {
         c[i] = new Complex();

       }
     }
     c[1].assign(0);

     theta4(ZERO, tau, EPS, c[0]);
     theta2(ZERO, tau, EPS, c[2]);
     theta3(ZERO, tau, EPS, c[3]);
   }

  /**
   * Returns theta constants &theta;<sub>0</sub>(0,&tau;), &theta;<sub>1</sub>(0,&tau;) = 0,
   * &theta;<sub>2</sub>(0,&tau;) = 0, and &theta;<sub>3</sub>(0,&tau;).
   * The result is stored in the array c.
   * The array may have null entries, which causes the routine to create
   * instances.
   * @param tau lattice paramter &tau;
   * @param eps absoute error for theta functions
   * @return array with the four theta constants (on output)
   */
  static Complex[] thetaConstants(final Complex tau,
                                         final double eps) {
    Complex[] c = new Complex[4];
    thetaConstants(tau, eps, c);
    return c;
  }

  /**
   * Returns theta constants &theta;<sub>0</sub>(0,&tau;), &theta;<sub>1</sub>(0,&tau;) = 0,
   * &theta;<sub>2</sub>(0,&tau;) = 0, and &theta;<sub>3</sub>(0,&tau;).
   * The result is stored in the array c.
   * The array may have null entries, which causes the routine to create
   * instances.
   * @param tau lattice paramter &tau;
   * @return array with the four theta constants (on output)
   */
  public static Complex[] thetaConstants(final Complex tau) {
    Complex[] c = new Complex[4];
    thetaConstants(tau, EPS, c);
    return c;
  }

  /**
   * Computes for given w and &tau; a representetive in
   * the lattice generated by 1 and &tau; such that w = z + m &pi; + n &pi; &tau;
   * and z = x &pi; + y &pi; &tau; with -0.5 <= x,y <= 0.5.
   * c is the quasi-periodic factor which gives:
   * &theta;<sub>3</sub>(w,&tau;) = c(z,&tau;,n) &theta;<sub>3</sub>(z,&tau;)
   * w and z may coinside.
   * @param w argument
   * @param tau lattice parameter
   * @param z = x &pi; + y &pi; &tau; with -0.5 <= x,y <= 0.5
   * @return period integer for &tau;
   */
  static int transformIntoFundamental(final Complex w, final Complex tau,
                                      final Complex z) {

    final double Y = (w.im) / tau.im / Math.PI; // Y = y+n;
    final double X = w.re / Math.PI - Y * tau.re; // X = x+m;

    final int m = (int) (Math.floor(X + 0.5));
    final int n = (int) (Math.floor(Y + 0.5));

    transform(w, tau, -m, -n, z);

    return n;
  }

  /**
   * Computes w = z + m &pi; + n &pi; &tau;
   * w and z may coinside.
   */
  static void transform(final Complex z, final Complex tau,
                        final int m, final int n,
                        final Complex w) {
    w.re = z.re + (m + n * tau.re) * Math.PI;
    w.im = z.im + (n * tau.im) * Math.PI;
  }

  /**
   * Computes w = z + x &pi; + y &pi; &tau;
   */
  static void transform(final Complex z, final Complex tau,
                        final double x, final double y,
                        final Complex w) {
    w.re = z.re + (x + y * tau.re) * Math.PI;
    w.im = z.im + (y * tau.im) * Math.PI;
  }

  /**
   * Computes integer N such that:
   *
   * | exp ( i &pi; &tau; N*N ) cos( 2NZ ) | < eps
   *
   * the term above can be used to estimate the error
   * of theta[34].
   *
   * @param z
   * @param tau
   * @param eps
   * @return N
   */
  static int lengthOfSeries(final Complex z, final Complex tau,
                            final double eps) {

    final double delta = -Math.PI * tau.im;
    final double y = Math.abs(z.im);

    return (int) Math.floor(1 -
                            (y + Math.sqrt(Math.log(eps) * delta + y * y)) /
                            delta);
  }

  /**
   * Periodicity factor for theta functions: c = exp (-i n (n &pi; &tau; + 2 z ) )
   *
   * theta[1234](w,&tau;) = (+/-) c(z,&tau;,n) &theta;<sub>4</sub>(z,&tau;)
   * with w = z + m &pi; + n &pi; &tau;.
   * Sign depends on theta function and on n and/or m.
   * @param z
   * @param tau
   * @param n
   * @param logOfC log(c)
   */
  static void logOfPeriodicityFactor(final Complex z, final Complex tau,
                                     final int n,
                                     final Complex logOfC) {

    logOfC.assign( -n * (n * Math.PI * tau.re + 2 * z.re),
                  -n * (n * Math.PI * tau.im + 2 * z.im));

    logOfC.assignTimesI();
  }

  static void logOfShiftFactor(final Complex z, final Complex tau,
                               final Complex logOfD) {
    logOfD.assign(Math.PI / 4 * tau.re + z.re,
                  Math.PI / 4 * tau.im + z.im);
    logOfD.assignTimesI();
  }

  static Complex logOfShiftFactor(final Complex z, Complex tau) {
    Complex logOfD = new Complex();
    logOfShiftFactor(z, tau, logOfD);
    return logOfD;
  }

  /* **** THETA **** */

  /**
   * Computes &theta;(z,&tau;).
   * &theta;, &theta;<sub>0</sub>, &theta;<sub>4</sub> refer to the same function, thus
   * {@link #theta(Complex,Complex)} ,
   * {@link #theta0(Complex,Complex)}, and
   * {@link #theta4(Complex,Complex)}
   * are identical.
   * @param z argument
   * @param tau lattice parameter
   * @return &theta;(z,&tau;)
   * @see #theta(Complex,Complex,Complex,Complex,Complex)
   */
  public static Complex theta(final Complex z, final Complex tau) {
    return theta4( z, tau );
  }

  /**
   * Helps to compute &theta;(w,&tau;) = c &theta;(z,&tau;).
   * &theta;, &theta;<sub>0</sub>, &theta;<sub>4</sub> refer to the same function, thus
   * {@link #theta(Complex,Complex,Complex,Complex,Complex)},
   * {@link #theta0(Complex,Complex,Complex,Complex,Complex)}, and
   * {@link #theta4(Complex,Complex,Complex,Complex,Complex)}
   * are idenical.
   *
   * @param w argument
   * @param tau lattice parameter
   * @param z = x &pi; + y &pi; &tau; with -0.5 <= x,y <= 0.5 (on output)
   * @param logOfC equals log(c) (on output)
   * @param thetaOfZ equals &theta;(z,&tau;) (on output)
   * @see #theta(Complex,Complex)
   */
  public static void theta(final Complex w, final Complex tau,
                           final Complex z,
                           final Complex logOfC,
                           final Complex thetaOfZ) {
    theta4(w, tau, EPS, z, logOfC, thetaOfZ);
  }

  /* **** THETA0 **** */

  /**
   * Computes &theta;<sub>0</sub>(z,&tau;).
   * &theta;, &theta;<sub>0</sub>, &theta;<sub>4</sub> refer to the same function, thus
   * {@link #theta(Complex,Complex)},
   * {@link #theta0(Complex,Complex)}, and
   * {@link #theta4(Complex,Complex)}
   * are identical.
   * @param z argument
   * @param tau lattice parameter
   * @return &theta;<sub>0</sub>(z,&tau;)
   * @see #theta0(Complex,Complex,Complex,Complex,Complex)
   */
  public static Complex theta0(final Complex z, final Complex tau) {
    return theta4(z,tau);
  }

  /**
   * Helps to compute &theta;<sub>0</sub>(w,&tau;) = c &theta;(z,&tau;).
   * &theta;, &theta;<sub>0</sub>, &theta;<sub>4</sub> refer to the same function, thus
   * {@link #theta(Complex,Complex,Complex,Complex,Complex)},
   * {@link #theta0(Complex,Complex,Complex,Complex,Complex)}, and
   * {@link #theta4(Complex,Complex,Complex,Complex,Complex)}
   * are identical.
   * @param w argument
   * @param tau lattice parameter
   * @param z = x &pi; + y &pi; &tau; with -0.5 <= x,y <= 0.5 (on output)
   * @param logOfC equals log(c) (on output)
   * @param thetaOfZ equals &theta;(z,&tau;) (on output)
   * @see #theta0(Complex,Complex)
   */
  public static void theta0(final Complex w, final Complex tau,
                            final Complex z,
                            final Complex logOfC,
                            final Complex thetaOfZ) {
    theta4(w, tau, EPS, z, logOfC, thetaOfZ);
  }

  /******  THETA 1 ******* */

  static void logOfPeriodicityFactorForTheta1(final Complex z,
                                              final Complex tau,
                                              final int m, final int n,
                                              final Complex logOfC) {

    logOfPeriodicityFactor(z, tau, n, logOfC);

    logOfC.im += (m + n) % 2 * Math.PI;

  }

  static void periodicityFactorForTheta1(final Complex z, final Complex tau,
                                         final int m, final int n,
                                         final Complex result) {
    logOfPeriodicityFactorForTheta1(z, tau, m, n, result);
    result.assignExp();
  }

  /**
   * Returns quasi-periodic factor.
   *
   * &theta;<sub>1</sub>(w,&tau;) = c(z,&tau;,m,n) &theta;<sub>1</sub>(z,&tau;)
   * with w = z + m &pi; + n &pi; &tau;.
   * @return c
   */
  static Complex periodicityFactorForTheta1(final Complex z, final Complex tau,
                                            final int m, final int n) {
    Complex result = new Complex();
    periodicityFactorForTheta1(z, tau, m, n, result);
    return result;
  }

  static Complex theta1(final Complex z, final Complex tau,
                        final double eps) {
    final Complex result = new Complex();
    theta1(z, tau, eps, result);
    return result;
  }

  /**
   * Computes &theta;<sub>1</sub>(z,&tau;).
   * @param z argument
   * @param tau lattice parameter
   * @return &theta;<sub>1</sub>(z,&tau;)
   * @see #theta1(Complex,Complex,Complex,Complex,Complex)
   */
  public static Complex theta1(final Complex z, final Complex tau) {
    final Complex result = new Complex();
    theta1(z, tau, EPS, result);
    return result;
  }

  static void theta1(final Complex z, final Complex tau,
                     final Complex theta1OfZ) {
    theta1(z, tau, EPS, theta1OfZ);
  }

  /**
   * Helps to compute &theta;<sub>1</sub>(w,&tau;) = c &theta;(z,&tau;).
   * @param w argument
   * @param tau lattice parameter
   * @param z equals x &pi; + y &pi; &tau; with -0.5 <= x,y <= 0.5 (on output)
   * @param logOfC equals log(c) (on output)
   * @param thetaOfZ equals &theta;(z,&tau;) (on output)
   * @see #theta1(Complex,Complex)
   */
  public static void theta1(final Complex w, final Complex tau,
                            final Complex z,
                            final Complex logOfC,
                            final Complex thetaOfZ) {
    theta1(w, tau, EPS, z, logOfC, thetaOfZ);
  }

  /**
   * Helps to compute &theta;<sub>1</sub>(w,&tau;) = c &theta;(z,&tau;).
   * @param w argument
   * @param tau lattice parameter
   * @param z equals x &pi; + y &pi; &tau; with -0.5 <= x,y <= 0.5 (on output)
   * @param logOfC equals log(c) (on output)
   * @param thetaOfZ equals &theta;(z,&tau;) (on output)
   * @see #theta1(Complex,Complex)
   */
  static void theta1(final Complex z, final Complex tau,
                     final double eps, final Complex thetaOfZ) {

    final Complex logOfC = new Complex();

    theta1(z, tau, eps, new Complex(), logOfC, thetaOfZ);

    logOfC.assignExp();

    thetaOfZ.assignTimes(logOfC);
  }

  /**
   * Helps to compute &theta;<sub>1</sub>(w,&tau;) = c &theta;(z,&tau;).
   * @param w argument
   * @param tau lattice parameter
   * @param eps abs error for theta4OfZ
   * @param z equals x &pi; + y &pi; &tau; with -0.5 <= x,y <= 0.5 (on output)
   * @param logOfC equals log(c) (on output)
   * @param thetaOfZ equals &theta;(z,&tau;) (on output)
   * @see #theta1(Complex,Complex)
   */
  static void theta1(final Complex w, final Complex tau,
                     final double eps,
                     final Complex z,
                     final Complex logOfC,
                     final Complex thetaOfZ) {

    transform(w, tau, 0.0, 0.5, z);

    theta4(z, tau, eps, z, logOfC, thetaOfZ);

    Complex logOfD = logOfShiftFactor(w, tau);

    logOfD.im -= 0.5 * Math.PI;

    logOfC.assignPlus(logOfD);
  }

  /******  THETA 2 ********/

  static void logOfPeriodicityFactorForTheta2(final Complex z,
                                              final Complex tau,
                                              final int m, final int n,
                                              final Complex logOfC) {

    logOfPeriodicityFactor(z, tau, n, logOfC);

    logOfC.im += m % 2 * Math.PI;

  }

  static void periodicityFactorForTheta2(final Complex z, final Complex tau,
                                         final int m, final int n,
                                         final Complex result) {
    logOfPeriodicityFactorForTheta2(z, tau, m, n, result);
    result.assignExp();
  }

  /**
   * Returns quasi-periodic factor.
   *
   * &theta;<sub>2</sub>(w,&tau;) = c(z,&tau;,m,n) &theta;<sub>2</sub>(z,&tau;)
   * with w = z + m &pi; + n &pi; &tau;.
   * @return c
   */
  static Complex periodicityFactorForTheta2(final Complex z, final Complex tau,
                                            final int m, final int n) {
    Complex result = new Complex();
    periodicityFactorForTheta2(z, tau, m, n, result);
    return result;
  }

  static Complex theta2(final Complex z, final Complex tau,
                        final double eps) {
    final Complex result = new Complex();
    theta2(z, tau, eps, result);
    return result;
  }


  /**
   * Computes &theta;<sub>2</sub>(z,&tau;).
   * @param z argument
   * @param tau lattice parameter
   * @return &theta;<sub>2</sub>(z,&tau;)
   * @see #theta2(Complex,Complex,Complex,Complex,Complex)
   */
  public static Complex theta2(final Complex z, final Complex tau) {
    final Complex result = new Complex();
    theta2(z, tau, EPS, result);
    return result;
  }

  static void theta2(final Complex z, final Complex tau,
                     final Complex theta2OfZ) {
    theta2(z, tau, EPS, theta2OfZ);
  }


  /**
   * Helps to compute &theta;<sub>2</sub>(w,&tau;) = c &theta;(z,&tau;).
   * @param w argument
   * @param tau lattice parameter
   * @param z equals x &pi; + y &pi; &tau; with -0.5 <= x,y <= 0.5 (on output)
   * @param logOfC equals log(c) (on output)
   * @param thetaOfZ equals &theta;(z,&tau;) (on output)
   * @see #theta2(Complex,Complex)
   */
  public static void theta2(final Complex w, final Complex tau,
                            final Complex z,
                            final Complex logOfC,
                            final Complex thetaOfZ) {
    theta2(w, tau, EPS, z, logOfC, thetaOfZ);
  }

  static void theta2(final Complex z, final Complex tau,
                            final double eps, final Complex thetaOfZ) {

    final Complex logOfC = new Complex();

    theta2(z, tau, eps, new Complex(), logOfC, thetaOfZ);

    logOfC.assignExp();

    thetaOfZ.assignTimes(logOfC);
  }

  /**
   * Helps to compute &theta;<sub>2</sub>(w,&tau;) = c &theta;<sub>4</sub>(z,&tau;).
   * @param w
   * @param tau
   * @param eps abs error for &theta;<sub>4</sub>OfZ
   * @param z = x &pi; + y &pi; &tau; with -0.5 <= x,y <= 0.5
   * @param logOfC = log(c)
   * @param theta4OfZ = &theta;(z,&tau;)
   */
  static void theta2(final Complex w, final Complex tau,
                     final double eps,
                     final Complex z,
                     final Complex logOfC,
                     final Complex thetaOfZ) {

    transform(w, tau, 0.5, 0.5, z);

    theta4(z, tau, eps, z, logOfC, thetaOfZ);

    Complex logOfD = logOfShiftFactor(w, tau);

    logOfC.assignPlus(logOfD);
  }

  /******  THETA 3 ********/

  static void logOfPeriodicityFactorForTheta3(final Complex z,
                                              final Complex tau,
                                              final int n,
                                              final Complex logOfC) {

    logOfPeriodicityFactor(z, tau, n, logOfC);
  }

  static void periodicityFactorForTheta3(final Complex z, final Complex tau,
                                         final int n,
                                         final Complex result) {
    logOfPeriodicityFactorForTheta3(z, tau, n, result);
    result.assignExp();
  }

  /**
   * Returns quasi-periodic factor.
   *
   * &theta;<sub>3</sub>(w,&tau;) = c(z,&tau;,n) &theta;<sub>3</sub>(z,&tau;)
   * with w = z + m &pi; + n &pi; &tau;.
   * @return c
   */
  static Complex periodicityFactorForTheta3(final Complex z, final Complex tau,
                                            final int n) {
    Complex result = new Complex();
    periodicityFactorForTheta3(z, tau, n, result);
    return result;
  }

  static Complex theta3(final Complex z, final Complex tau,
                        final double eps) {
    final Complex result = new Complex();
    theta3(z, tau, eps, result);
    return result;
  }

  /**
   * Computes &theta;<sub>3</sub>(z,&tau;).
   * @param z argument
   * @param tau lattice parameter
   * @return &theta;<sub>3</sub>(z,&tau;)
   * @see #theta3(Complex,Complex,Complex,Complex,Complex)
   */
  public static Complex theta3(final Complex z, final Complex tau) {
    final Complex result = new Complex();
    theta3(z, tau, EPS, result);
    return result;
  }

  static void theta3(final Complex z, final Complex tau,
                     final Complex theta3OfZ) {
    theta3(z, tau, EPS, theta3OfZ);
  }

  static void theta3(final Complex z, final Complex tau,
                     final double eps, final Complex theta3OfZ) {

    final Complex logOfC = new Complex();

    theta3(z, tau, eps, new Complex(), logOfC, theta3OfZ);

    logOfC.assignExp();

    theta3OfZ.assignTimes(logOfC);
  }

  /**
   * Helps to compute &theta;<sub>3</sub>(w,&tau;) = c &theta;(z,&tau;).
   * @param w argument
   * @param tau lattice parameter
   * @param z equals x &pi; + y &pi; &tau; with -0.5 <= x,y <= 0.5 (on output)
   * @param logOfC equals log(c) (on output)
   * @param thetaOfZ equals &theta;(z,&tau;) (on output)
   * @see #theta3(Complex,Complex)
   */
  public static void theta3(final Complex w, final Complex tau,
                            final Complex z,
                            final Complex logOfC,
                            final Complex thetaOfZ) {
    theta3(w, tau, EPS, z, logOfC, thetaOfZ);
  }

  /**
   * Helps to compute &theta;<sub>3</sub>(w,&tau;) = c &theta;<sub>4</sub>(z,&tau;).
   * @param w
   * @param tau
   * @param eps abs error for theta4OfZ
   * @param z = x &pi; + y &pi; &tau; with -0.5 <= x,y <= 0.5
   * @param logOfC =log(c)
   * @param theta4OfZ = &theta;<sub>4</sub>(z,&tau;)
   */
  static void theta3(final Complex w, final Complex tau,
                            final double eps,
                            final Complex z,
                            final Complex logOfC,
                            final Complex thetaOfZ) {

    transform(w, tau, 0.5, 0, z);

    theta4(z, tau, eps, z, logOfC, thetaOfZ);
  }

  /* *****  THETA4 ******* */

  static void logOfPeriodicityFactorForTheta4
      (final Complex z, final Complex tau,
       final int n,
       final Complex logOfC) {

    logOfPeriodicityFactor(z, tau, n, logOfC);

    logOfC.im += (n % 2) * Math.PI;
  }

  static void periodicityFactorForTheta4
      (final Complex z, final Complex tau,
       final int n,
       final Complex result) {
    logOfPeriodicityFactorForTheta4(z, tau, n, result);
    result.assignExp();
  }

  /**
   * Returns quasi-periodic factor.
   *
   * &theta;<sub>4</sub>(w,&tau;) = c(z,&tau;,n) &theta;(z,&tau;)
   * with w = z + m &pi; + n &pi; &tau;.
   * @return c
   */
  static Complex periodicityFactorForTheta4(final Complex z, final Complex tau,
                                            final int n) {
    Complex result = new Complex();
    periodicityFactorForTheta4(z, tau, n, result);
    return result;
  }

  static Complex theta4(final Complex z, final Complex tau,
                        final double eps) {
    final Complex result = new Complex();
    theta4(z, tau, eps, result);
    return result;
  }

  static void theta4(final Complex z, final Complex tau,
                     final Complex theta4OfZ) {
    theta4(z, tau, EPS, theta4OfZ);
  }

  static void theta4(final Complex z, final Complex tau,
                     final double eps, final Complex theta4OfZ) {

    final Complex logOfC = new Complex();

    theta4(z, tau, eps, new Complex(), logOfC, theta4OfZ);

    logOfC.assignExp();

    theta4OfZ.assignTimes(logOfC);
  }

  /**
   * Computes &theta;<sub>4</sub>(w,&tau;).
   * &theta;, &theta;<sub>0</sub>, &theta;<sub>4</sub> refer to the same function, thus
   * {@link #theta(Complex,Complex)},
   * {@link #theta0(Complex,Complex)}, and
   * {@link #theta4(Complex,Complex)}
   * are identical.
   * @param z argument
   * @param tau lattice parameter
   * @see #theta4(Complex,Complex,Complex,Complex,Complex)
   */
  public static Complex theta4(final Complex z, final Complex tau) {
    final Complex result = new Complex();
    theta4(z, tau, EPS, result);
    return result;
  }

  /**
   * Helps to compute &theta;<sub>4</sub>(w,&tau;) = c &theta;(z,&tau;).
   * &theta;, &theta;<sub>0</sub>, &theta;<sub>4</sub> refer to the same function, thus
   * {@link #theta(Complex,Complex,Complex,Complex,Complex)},
   * {@link #theta0(Complex,Complex,Complex,Complex,Complex)}, and
   * {@link #theta4(Complex,Complex,Complex,Complex,Complex)}
   * are identical.
   * @param w argument
   * @param tau lattice parameter
   * @param z = x &pi; + y &pi; &tau; with -0.5 <= x,y <= 0.5 (on output)
   * @param logOfC =log(c) (on output)
   * @param thetaOfZ = &theta;<sub>4</sub>(z,&tau;) (on output)
   * @see #theta4(Complex,Complex)
   */
  public static void theta4(final Complex w, final Complex tau,
                            final Complex z,
                            final Complex logOfC,
                            final Complex thetaOfZ) {
    theta4(w, tau, EPS, z, logOfC, thetaOfZ);
  }

  /**
   * Helps to compute &theta;<sub>4</sub>(w,&tau;) = c &theta;(z,&tau;).
   * &theta;, &theta;<sub>0</sub>, &theta;<sub>4</sub> refer to the same function.
   * @param w argument
   * @param tau lattice paramter &tau;
   * @param eps abs error for thetaOfZ
   * @param z = x &pi; + y &pi; &tau; with -0.5 <= x,y <= 0.5 (on output)
   * @param logOfC equals log(c) (on output)
   * @param thetaOfZ equals &theta;(z,&tau;) (on output)
   */
  static void theta4(final Complex w, final Complex tau,
                     final double eps,
                     final Complex z,
                     final Complex logOfC,
                     final Complex thetaOfZ) {

    final int n = transformIntoFundamental(w, tau, z);
    logOfPeriodicityFactorForTheta4(z, tau, n, logOfC);
    theta4PlainSeries(z, tau, eps, thetaOfZ);
  }

  static void theta4PlainSeries(final Complex z, final Complex tau,
                                final double eps, final Complex result) {

    if (! (tau.im > 0)) {
      throw new IllegalArgumentException(
          "imagiary part of tau must be positive");
    }

    final Complex tauIPi = tau.times(new Complex(0, Math.PI));

    final Complex tauIPiNSqr = new Complex();

    final Complex iz = z.timesI();

    final Complex iZ2N = new Complex();
    final Complex term = new Complex();
    final Complex term1 = new Complex();
    final Complex term2 = new Complex();

    final int N = lengthOfSeries(z, tau, eps);

    double sign = -1;

    result.assign(1);

    for (int n = 1; n <= N; n++, sign *= -1) {

      iZ2N.assignTimes(iz, 2 * n);

      tauIPiNSqr.assignTimes(tauIPi, n * n);

      term1.assignPlus(tauIPiNSqr, iZ2N);
      term1.assignExp();

      term2.assignMinus(tauIPiNSqr, iZ2N);
      term2.assignExp();

      term.assignPlus(term1, term2);
      term.assignTimes(sign);

      result.assignPlus(term);
    }
  }

}
