/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2003-2010, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.ellipticFunctions;

import de.jtem.mfc.field.Complex;

/**
 * Implementation for Jacobi's elliptic functions.
 *
 * Attention! This class is not thread save.
 * This is usually no problem, when you do numerics.
 *
 * @author Markus Schmies, Boris Springborn,
 */

public class Jacobi
    extends Elliptic {

  Jacobi() {
  }

  final static Complex z = new Complex();
  final static Complex u = new Complex();
  final static Complex Z = new Complex();
  final static Complex c = new Complex();
  final static Complex g = new Complex();
  final static Complex m = new Complex();

  final static Complex k = new Complex();
  final static Complex K = new Complex();

  final static Complex kPrime = new Complex();
  final static Complex KPrime = new Complex();

  final static Complex tau = new Complex();

  final static Complex an = new Complex();
  final static Complex bn = new Complex();

  final static Complex ONE = new Complex(1);

  final static Complex logOfC = new Complex();

  final static Complex theta1 = new Complex();
  final static Complex theta2 = new Complex();
  final static Complex theta3 = new Complex();
  final static Complex theta4 = new Complex();

  final static Complex snOfU = new Complex();
  final static Complex cnOfU = new Complex();
  final static Complex dnOfU = new Complex();

  final static Complex [] thetaConstants = new Complex[4];

  final static Complex logOfCForTheta1 = new Complex();
  final static Complex logOfCForTheta2 = new Complex();
  final static Complex logOfCForTheta3 = new Complex();
  final static Complex logOfCForTheta4 = new Complex();

    /**
     * Computes sn(u,k(&tau;)) for given latiuce paramter &tau;, and theta constants.
     * @param u argument
     * @param tau lattice parameter
     * @param thetaConstants theta constants
     * @param snOfU sn(u,k(&tau;)) (on output)
     * @see #thetaConstants(Complex)
     * @see #tau_from_k(Complex)
     * @see #sn(Complex,Complex)
     * @see #sn(double,double)
     */
    public static void sn(final Complex u, final Complex tau,
                          final Complex[] thetaConstants,
                          final Complex snOfU) {

      z.assignDivide(u, thetaConstants[3]);
      z.assignDivide(thetaConstants[3]);

      theta1(z, tau, EPS, Z, logOfCForTheta1, theta1);
      theta4(z, tau, EPS, Z, logOfCForTheta4, theta4);

      logOfC.assignMinus(logOfCForTheta1, logOfCForTheta4);

      snOfU.assignExp(logOfC);

      snOfU.assignDivide(thetaConstants[2]);
      snOfU.assignTimes(thetaConstants[3]);
      snOfU.assignTimes(theta1);
      snOfU.assignDivide(theta4);
    }

    /**
     * Returns sn(u,k). It computes the lattice paramter
     * &tau; as function of k, and further the @theta;-constants.
     * Therefore in most cases it is more efficient to hash
     * these values and call
     * {@link #sn(Complex,Complex,Complex[],Complex)}.
     * @param u argument
     * @param k modulus
     * @return sn(u,k)
     * @see #thetaConstants(Complex)
     * @see #tau_from_k(Complex)
     * @see #sn(Complex,Complex,Complex[],Complex)
     * @see #sn(double,double)
     */
    public static Complex sn(final Complex u, final Complex k) {

      tau_from_k(k, tau);

      thetaConstants(tau, EPS, thetaConstants);

      final Complex snOfU = new Complex();

      sn(u, tau, thetaConstants, snOfU);

      return snOfU;
    }

    /**
     * Returns sn(u,k). It computes the lattice paramter
     * &tau; as function of k, and further the @theta;-constants.
     * Therefore in most cases it is more efficient to hash
     * these values and call
     * {@link #sn(Complex,Complex,Complex[],Complex)}.
     * @param u argument
     * @param k modulus
     * @return sn(u,k)
     * @see #thetaConstants(Complex)
     * @see #tau_from_k(Complex)
     * @see #sn(Complex,Complex,Complex[],Complex)
     * @see #sn(Complex,Complex)
     */
    public static double sn(final double u, final double k) {

      Jacobi.k.assign(k);
      Jacobi.u.assign(u);

      tau_from_k(Jacobi.k, tau);

      thetaConstants(tau, EPS, thetaConstants);

      sn(Jacobi.u, tau, thetaConstants, snOfU);

      return snOfU.re;
    }

    /**
     * Computes cn(u,k(&tau;)) for given latiuce paramter &tau;, and theta constants.
     * @param u argument
     * @param tau lattice parameter
     * @param thetaConstants theta constants
     * @param cnOfU sn(u,k(&tau;)) (on output)
     * @see #thetaConstants(Complex)
     * @see #tau_from_k(Complex)
     * @see #cn(Complex,Complex)
     */
    public static void cn(final Complex u, final Complex tau,
                          final Complex[] thetaConstants,
                          final Complex cnOfU) {

      z.assignDivide(u, thetaConstants[3]);
      z.assignDivide(thetaConstants[3]);

      theta2(z, tau, EPS, Z, logOfCForTheta2, theta2);
      theta4(z, tau, EPS, Z, logOfCForTheta4, theta4);

      logOfC.assignMinus(logOfCForTheta2, logOfCForTheta4);

      cnOfU.assignExp(logOfC);

      cnOfU.assignDivide(thetaConstants[2]);
      cnOfU.assignTimes(thetaConstants[0]);
      cnOfU.assignTimes(theta2);
      cnOfU.assignDivide(theta4);
    }

    /**
     * Returns cn(u,k). It computes the lattice paramter
     * &tau; as function of k, and further the @theta;-constants.
     * Therefore in most cases it is more efficient to hash
     * these values and call
     * {@link #cn(Complex,Complex,Complex[],Complex)}.
     * @param u argument
     * @param k modulus
     * @return cn(u,k)
     * @see #thetaConstants(Complex)
     * @see #tau_from_k(Complex)
     * @see #cn(Complex,Complex,Complex[],Complex)
     */
    public static Complex cn(final Complex u, final Complex k) {

      tau_from_k(k, tau);

      thetaConstants(tau, EPS, thetaConstants);

      final Complex cnOfU = new Complex();

      cn(u, tau, thetaConstants, cnOfU);

      return cnOfU;
    }

    /**
     * Returns cn(u,k). It computes the lattice paramter
     * &tau; as function of k, and further the @theta;-constants.
     * Therefore in most cases it is more efficient to hash
     * these values and call
     * {@link #cn(Complex,Complex,Complex[],Complex)}.
     * @param u argument
     * @param k modulus
     * @return cn(u,k)
     * @see #thetaConstants(Complex)
     * @see #tau_from_k(Complex)
     * @see #cn(Complex,Complex,Complex[],Complex)
     * @see #cn(Complex,Complex)
     */
    public static double cn(final double u, final double k) {

      Jacobi.k.assign(k);
      Jacobi.u.assign(u);

      tau_from_k(Jacobi.k, tau);

      thetaConstants(tau, EPS, thetaConstants);

      cn(Jacobi.u, tau, thetaConstants, cnOfU);

      return cnOfU.re;
    }

    /**
     * Computes dn(u,k(&tau;)) for given latiuce paramter &tau;, and theta constants.
     * @param u argument
     * @param tau lattice parameter
     * @param thetaConstants theta constants
     * @param dnOfU sn(u,k(&tau;)) (on output)
     * @see #thetaConstants(Complex)
     * @see #tau_from_k(Complex)
     * @see #dn(Complex,Complex)
     */
    public static void dn(final Complex u, final Complex tau,
                          final Complex[] thetaConstants,
                          final Complex dnOfU) {

      z.assignDivide(u, thetaConstants[3]);
      z.assignDivide(thetaConstants[3]);

      theta3(z, tau, EPS, Z, logOfCForTheta3, theta3);
      theta4(z, tau, EPS, Z, logOfCForTheta4, theta4);

      logOfC.assignMinus(logOfCForTheta3, logOfCForTheta4);

      dnOfU.assignExp(logOfC);

      dnOfU.assignDivide(thetaConstants[3]);
      dnOfU.assignTimes(thetaConstants[0]);
      dnOfU.assignTimes(theta3);
      dnOfU.assignDivide(theta4);
    }

    /**
     * Returns dn(u,k). It computes the lattice paramter
     * &tau; as function of k, and further the @theta;-constants.
     * Therefore in most cases it is more efficient to hash
     * these values and call
     * {@link #dn(Complex,Complex,Complex[],Complex)}.
     * @param u argument
     * @param k modulus
     * @return dn(u,k)
     * @see #thetaConstants(Complex)
     * @see #tau_from_k(Complex)
     * @see #dn(Complex,Complex,Complex[],Complex)
     */
    public static Complex dn(final Complex u, final Complex k) {

      tau_from_k(k, tau);

      thetaConstants(tau, EPS, thetaConstants);

      final Complex dnOfU = new Complex();

      dn(u, tau, thetaConstants, dnOfU);

      return dnOfU;
    }


    /**
     * Returns dn(u,k). It computes the lattice paramter
     * &tau; as function of k, and further the @theta;-constants.
     * Therefore in most cases it is more efficient to hash
     * these values and call
     * {@link #dn(Complex,Complex,Complex[],Complex)}.
     * @param u argument
     * @param k modulus
     * @return dn(u,k)
     * @see #thetaConstants(Complex)
     * @see #tau_from_k(Complex)
     * @see #dn(Complex,Complex,Complex[],Complex)
     * @see #dn(Complex,Complex)
     */
    public static double dn(final double u, final double k) {

      Jacobi.k.assign(k);
      Jacobi.u.assign(u);

      tau_from_k(Jacobi.k, tau);

      thetaConstants(tau, EPS, thetaConstants);

      dn(Jacobi.u, tau, thetaConstants, dnOfU);

      return dnOfU.re;
    }

  /**
   * Computes complementary modulus k' as function of the modulus.
   * @param k modulus
   * @param kPrime complementary modulus k'= sqrt( 1-k*k ) (on output)
   * @see #kPrime_from_k(Complex)
   */
  public static void kPrime_from_k(final Complex k, final Complex kPrime) {
    kPrime.assignSqr(k);
    kPrime.assignNeg();
    kPrime.assignPlus(ONE);
    kPrime.assignSqrt();
  }

  /**
   * Returns complementary modulus k' as function of the modulus.
   * @param k modulus
   * @return complementary modulus k'= sqrt( 1-k*k )
   * @see #kPrime_from_k(Complex,Complex)
   */
  public static Complex kPrime_from_k(final Complex k) {
    final Complex kPrime = new Complex();
    kPrime_from_k(k, kPrime);
    return kPrime;
  }

  /**
    * Returns complementary modulus k' as function of the modulus.
    * @param k modulus
    * @return complementary modulus k'= sqrt( 1-k*k )
    * @see #kPrime_from_k(Complex,Complex)
    */
   public static double kPrime_from_k(final double k) {
     return Math.sqrt( 1-k*k );
   }

  /**
   * Computes  m = a(infty) = b(infty)
   * pair series a(n+1) = ( a(n)+b(n) )/2, b(n+1) = sqrt(a(n)b(n))
   * <br>
   * see D.F.Lawden, Elliptic Functinos and Applications, page 81
   * 
   * @param a a(0)
   * @param b b(0)
   * @param eps relative error
   * @param m a(infty) = b(infty)
   */
  static void arithmeticGeometricMeanAlgorigthm
      (final Complex a, final Complex b,
       final double eps,
       final Complex m) {

    an.assign(a);
    bn.assign(b);

    for (int i = 0; i < 100; i++) {

      m.re = (an.re + bn.re) / 2;
      m.im = (an.im + bn.im) / 2;

      if (an.distSqr(bn) < an.absSqr() * eps * eps) {
        return;
      }

      g.assignTimes(an, bn);
      bn.assignSqrt(g);
      an.assign(m);
    }

    throw new RuntimeException("reached max count; algorithm failed");
  }

  /**
   * Computes quarter period K as function of the modulus k.
   * @param k modulus
   * @param K quarter period K (on output)
   * #see K_from_k(Complex)
   */
  public static void K_from_k(final Complex k, final Complex K) {

    kPrime_from_k(k, kPrime);

    arithmeticGeometricMeanAlgorigthm(ONE, kPrime, EPS, m);

    K.assign(Math.PI / 2);
    K.assignDivide(m);
  }

  /**
   * Returns quarter period K as function of the modulus k.
   * @param k modulus
   * @return quarter period
   * #see K_from_k(Complex,Complex)
   */
  public static Complex K_from_k(final Complex k) {
    final Complex K = new Complex();
    K_from_k(k, K);
    return K;
  }

  /**
   * Computes cmplementary quarter period K' as function of the modulus k.
   * @param k modulus
   * @param KPrime complementary quarter period K' (on output)
   * @see #KPrime_from_k(Complex)
   */
  public static void KPrime_from_k(final Complex k, final Complex KPrime) {

    arithmeticGeometricMeanAlgorigthm(ONE, k, EPS, m);

    KPrime.assign(Math.PI / 2);
    KPrime.assignDivide(m);
  }

  /**
   * Returns complementary quarter Period K' as function of the modulus k.
   * @param k modulus
   * @return complementary quarter Period K'
   * @see #KPrime_from_k(Complex,Complex)
   */
  public static Complex KPrime_from_k(final Complex k) {
    final Complex KPrime = new Complex();
    KPrime_from_k(k, KPrime);
    return KPrime;
  }

  /**
   * Computes modulus k as function of theta constants
   * @param thetaConstants theta constants
   * @param k modulus (on output)
   * @see #thetaConstants(Complex)
   * @see #k_from_thetaConstants(Complex[])
   */
  public static void k_from_thetaConstants(
      final Complex [] thetaConstants,
      final Complex k) {

    k.assignDivide(thetaConstants[2], thetaConstants[3]);
    k.assignSqr();
  }

  /**
   * Returns modulus k as function of theta constants.
   * @param thetaConstants theta constants
   * @return modulus
   * @see #thetaConstants(Complex)
   * @see #k_from_thetaConstants(Complex[],Complex)
   */
  public static Complex k_from_thetaConstants( final Complex [] thetaConstants ) {
    final Complex k = new Complex();
    k_from_thetaConstants(thetaConstants, k);
    return k;
  }

  /**
   * Computes complementary modulus k' as function theta constants.
   * @param thetaConstants theta constants
   * @param kPrime complementary modulus k' (on output)
   * @see #thetaConstants(Complex)
   * @see #kPrime_from_thetaConstants(Complex[])
   */
  public static void kPrime_from_thetaConstants(
      final Complex [] thetaConstants,
      final Complex kPrime) {

    kPrime.assignDivide(thetaConstants[0], thetaConstants[3]);
    kPrime.assignSqr();
  }

  /**
   * Returns complementary modulus k' as function of theta constants.
   * @param thetaConstants theta constants
   * @return complementary modulus
   * @see #thetaConstants(Complex)
   * @see #kPrime_from_thetaConstants(Complex[],Complex)
   */
  public static Complex kPrime_from_thetaConstants(
      final Complex [] thetaConstants ) {
    final Complex kPrime = new Complex();
    kPrime_from_thetaConstants(thetaConstants, kPrime);
    return kPrime;
  }

  /**
   * Computes quarter period K as function of theta constants.
   * @param thetaConstants theta constants
   * @param K quarter period (on output)
   * @see #thetaConstants(Complex)
   * @see #K_from_thetaConstants(Complex[])
   */
  public static void K_from_thetaConstants(
      final Complex [] thetaConstants, final Complex K) {

    K.assignSqr(thetaConstants[3]);
    K.assignTimes(Math.PI / 2);
  }

  /**
   * Returns quarter period K as function of theta constants.
   * @param thetaConstants theta constants
   * @return quarter period
   * @see #thetaConstants(Complex)
   * @see #K_from_thetaConstants(Complex[],Complex)
   */
  public static Complex K_from_thetaConstants(
      final Complex [] thetaConstants ) {
    final Complex K = new Complex();
    K_from_thetaConstants(thetaConstants, K);
    return K;
  }

  /**
   * Computes complementary quarter period K' as function of
   * quarter period K and lattice parameter &tau;.
   * @param K quarter period
   * @param tau lattice parameter
   * @param KPrime complementary quarter period (on output)
   * @see #KPrime_from_K_and_tau(Complex,Complex)
   */
  public static void KPrime_from_K_and_tau(
      final Complex K, final Complex tau, final Complex KPrime) {
    KPrime.assignTimes(tau, K);
    KPrime.assignDivideI();
  }

  /**
   * Returns complementary quarter period K' as function of
   * quarter period K and lattice parameter &tau;.
   * @param K quarter period
   * @param tau lattice parameter
   * @return complementary quarter period
   * @see #KPrime_from_K_and_tau(Complex,Complex,Complex)
   */
  public static Complex KPrime_from_K_and_tau(
      final Complex K, final Complex tau) {
    final Complex KPrime = new Complex();
    KPrime_from_K_and_tau( K, tau, KPrime );
    return KPrime;
  }

    /**
     * Computes quarter period K as function of
     * complementary quarter period K' and lattice parameter &tau;.
     * @param KPrime complementary quarter period
     * @param tau lattice parameter
     * @param K quarter period (on output)
     * @see #K_from_KPrime_and_tau(Complex,Complex)
     */
    public static void K_from_KPrime_and_tau(
        final Complex KPrime, final Complex tau, final Complex K) {
      K.assignDivide(KPrime,tau);
      K.assignTimesI();
    }

    /**
     * Returns quarter period K as function of
     * complementary quarter period K' and lattice parameter &tau;.
     * @param KPrime complementary quarter period
     * @param tau lattice parameter
     * @return quarter period (on output)
     * @see #K_from_KPrime_and_tau(Complex,Complex,Complex)
     */
    public static Complex K_from_KPrime_and_tau(
        final Complex KPrime, final Complex tau) {
      final Complex K = new Complex();
      K_from_KPrime_and_tau( KPrime, tau, K );
      return K;
    }

    /**
     * Computes complementary lattice paramter &tau;' = -1/&tau; for lattice
     * parameter &tau;.
     * tau and tauPrime may coinside.
     * @param tau lattice paramter &tau;
     * @param tauPrime complementary lattice paramter &tau;'(on output)
     * @see #tauPrime_from_tau(Complex)
     */
    public static void tauPrime_from_tau( final Complex tau,
                                          final Complex tauPrime ) {

      tauPrime.assignInvert(tau);
      tauPrime.assignNeg();
    }

    /**
     * Returns complementary lattice paramter &tau;' = -1/&tau; for lattice
     * parameter &tau;.
     * @param tau lattice paramter &tau;
     * @return complementary lattice paramter &tau;'
     * @see #tauPrime_from_tau(Complex,Complex)
     */
    public static Complex tauPrime_from_tau( final Complex tau ) {
      final Complex tauPrime = new Complex();
      tauPrime_from_tau( tau, tauPrime );
      return tauPrime;
    }

    /**
     * Computes lattice parameter &tau; as function of modulus k.
     * @param k modulus
     * @param tau lattice paramter &tau;
     * @see #tau_from_k(Complex)
     */
    public static void tau_from_k(final Complex k, final Complex tau) {

      Jacobi.KPrime_from_k(k, KPrime);
      Jacobi.K_from_k(k, K);

      tau.assignDivide(KPrime, K);
      tau.assignTimesI();
    }


    /**
     * Computes lattice parameter &tau; as function of modulus k.
     * @param k modulus
     * @return lattice paramter
     * @see #tau_from_k(Complex,Complex)
     */
    public static Complex tau_from_k( final Complex k ) {
      final Complex tau = new Complex();
      tau_from_k( k, tau );
      return tau;
    }



}
